#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>

struct account_info {
  char name[1024];
  unsigned long long value;
};

static char *read_hostfile(int fd, size_t *bytes)
{
  size_t readed = 0;
  size_t allocated = 0;
  char *lines = NULL;
  char *newlines;
  ssize_t ret;

  while (1) {
    if (allocated == readed) {
      allocated = (allocated == 0 ? 4096 : allocated * 2);
      newlines = realloc(lines, allocated);
      if (newlines == NULL)
	goto error;
      lines = newlines;
    }
    ret = read(fd, &lines[readed], allocated - readed);
    if (ret < 0) {
      if (errno == EAGAIN || errno == EINTR)
	continue;
      fprintf(stderr, "read error on hostfile: %s\n", strerror(errno));
      goto error;
    } else if (ret == 0) {
      break;
    }
    readed += ret;
  }

  *bytes = readed;
  return lines;

 error:
  free(lines);
  return NULL;
}


static ssize_t get_line(char *dst, size_t dstlen, 
			const char *src, size_t offs, size_t maxoffs)
{
  size_t ind = 0;

  while ((offs + ind) < maxoffs) {
    if (src[offs + ind] == '\n' || src[offs + ind] == 0) {
      if (ind >= dstlen)
	goto error;
      strncpy(dst, &src[offs], ind);
      dst[ind] = 0;
      return offs + ind + 1;
    }
    ind++;
  }

  if (ind > 0) {
    if (ind >= dstlen)
      goto error;
    strncpy(dst, &src[offs], ind);
    dst[ind] = 0;
    return offs + ind;
  }

 error:
  return -1;
}

static int read_account_info(struct account_info **acc, int fd)
{
  char host[1024];
  ssize_t new_offs;
  size_t firstspace;
  char *lines;
  size_t bytes;
  size_t offs;
  int naccs = 0;
  struct account_info curacc;
  char *endptr;

  *acc = NULL;

  lines = read_hostfile(fd, &bytes);
  if (lines == NULL) {
    fprintf(stderr, "could not read hostfile\n");
    exit(-1);
  }

  offs = 0;
  while (offs < bytes) {

    new_offs = get_line(host, sizeof(host), lines, offs, bytes);
    if (new_offs < 0)
      break;
    if ((size_t) new_offs == offs) {
      fprintf(stderr, "bug: new_offs == offs\n");
      exit(-1);
    }
    offs = new_offs;

    if (host[0] == '#')
      continue;

    if (strlen(host) == 0)
      continue;

    /* get number of bytes accounted. assume zero bytes when information
       is not available in the file */
    firstspace = 0;
    while (!isspace(host[firstspace]) && host[firstspace] != 0)
      firstspace++;

    curacc.value = 0;

    strncpy(curacc.name, host, firstspace);
    curacc.name[firstspace] = 0;

    firstspace++;
    curacc.value = strtoull(&host[firstspace], &endptr, 10);
    if (*endptr != 0) {
      fprintf(stderr, "illegal number in host file: %s\n", host);
      exit(-1);
    }

    *acc = realloc(*acc, sizeof(struct account_info) * (naccs + 1));
    if (*acc == NULL) {
      fprintf(stderr, "out of memory for accounting info\n");
      exit(-1);
    }
    (*acc)[naccs] = curacc;
    naccs++;
  }

  free(lines);
  return naccs;
}


void get_account(char *dsthost, size_t hostmaxlen, const char *hostfile)
{
  int fd;
  unsigned long long smallest = 0;
  int smallest_i = -1;
  struct account_info *acc;
  int nacc;
  int i;

  fd = open(hostfile, O_RDWR);
  if (fd < 0) {
    fprintf(stderr, "can not open hostfile %s: %s\n", hostfile, strerror(errno));
    exit(-1);
  }

  while (lockf(fd, F_TLOCK, 0)) {
    fprintf(stderr, "lockfile: %s. sleeping a second.\n", strerror(errno));
    sleep(1);
  }

  nacc = read_account_info(&acc, fd);
  if (nacc == 0) {
    fprintf(stderr, "no hosts found from %s\n", hostfile);
    exit(-1);
  }

  for (i = 0; i < nacc; i++) {
    if (smallest_i < 0)
      smallest = acc[i].value + 1;
    if (acc[i].value < smallest) {
      smallest = acc[i].value;
      smallest_i = i;
    }
  }

  if (smallest_i < 0) {
    fprintf(stderr, "no hosts found from %s\n", hostfile);
    exit(-1);
  }

  fprintf(stderr, "%s has smallest number of bytes: %llu\n", acc[smallest_i].name, acc[smallest_i].value);

  strncpy(dsthost, acc[smallest_i].name, hostmaxlen);
  dsthost[hostmaxlen] = 0 ;

  free(acc);

  lockf(fd, F_ULOCK, 0);
  close(fd);
}

static ssize_t full_write(int fd, void *buf, size_t length)
{
  char *wbuf = buf;
  ssize_t ret;
  size_t written = 0;
  while (written < length) {
    ret = write(fd, &wbuf[written], length - written);
    if (ret < 0) {
      if (errno == EAGAIN || errno == EINTR)
	continue;
      break;
    } else if (ret == 0) {
      fprintf(stderr, "interesting, write returned zero\n");
      break;
    }
    written += ret;
  }
  return written;
}

void write_account_info(struct account_info *acc, int nacc, int fd)
{
  int i;
  char host[1024];
  for (i = 0; i < nacc; i++) {
    snprintf(host, sizeof(host), "%s %llu\n", acc[i].name, acc[i].value);
    full_write(fd, host, strlen(host));
  }
}

void account_increment(char *dsthost, const char *hostfile,
		       unsigned long long incr)
{
  int fd;
  struct account_info *acc;
  int nacc;
  int i;

  fd = open(hostfile, O_RDWR);
  if (fd < 0) {
    fprintf(stderr, "can not open hostfile %s: %s\n", hostfile, strerror(errno));
    exit(-1);
  }

  while (lockf(fd, F_TLOCK, 0)) {
    fprintf(stderr, "lockfile: %s. sleeping a second.\n", strerror(errno));
    sleep(1);
  }

  nacc = read_account_info(&acc, fd);
  if (nacc == 0) {
    fprintf(stderr, "account info not found\n");
    return;
  }

  for (i = 0; i < nacc; i++) {
    if (strcmp(dsthost, acc[i].name) == 0) {
      acc[i].value += incr;
      fprintf(stderr, "incrementing %s to %llu\n", acc[i].name, acc[i].value);
      if (ftruncate(fd, 0)) {
	fprintf(stderr, "can not truncate target list file: %s\n", strerror(errno));
	exit(-1);
      }
      lseek(fd, 0, SEEK_SET);
      write_account_info(acc, nacc, fd);
      break;
    }
  }

  free(acc);

  lockf(fd, F_ULOCK, 0);
  close(fd);
}
