/* connect two port forward utility
Copyright (c) 2003 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"

extern int errno;

/* configuration variables */
static int default_interval = 2000;

static int hello_from_src = 0;

/* per-client variables */
static pid_t cpid;

static void pipemeter_sigpipe(int sig)
{
  sig = sig;
  fprintf(stderr, "%d: connecttwo: SIGPIPE caught. Killing client.\n", getpid());
}

struct pipe {
  int src_fd;
  int dst_fd;
  char *buf;
  int bufsize;
  int bytes;
  int offs;
  int eof_read;
  long long tbytes;
  long long wbytes;
  struct timeval ot;
  struct timeval vot;
};

static int timetest(char *s, struct pipe *p, int force)
{
  struct timeval nt;
  int t;
  strcpy(s, "bandwidth: NaN");
  if (gettimeofday(&nt, 0)) {
    /* time failed. bandwidth = NaN. return false. */
    return 0;
  }
  t = 1000 * (nt.tv_sec - p->ot.tv_sec) + ((int) nt.tv_usec)/1000 - ((int) p->ot.tv_usec)/1000;
  if (t < 0) {
    fprintf(stderr, "connecttwo: chronoton particles detected. clock ran backwards. k3wl!\n");
    t = default_interval + 1;
  }
  if (t > default_interval || force) {
    int order;
    double bw;
    char id[256];
    if (t) {
      bw = ((double) p->wbytes) * 1000.0f / ((double) t);
      order = 0;
      while (bw >= 1024.0f) {
	bw /= 1024.0f;
	order++;
	if (order == 4)
	  break;
      }
      switch (order) {
      case 0:
	strcpy(id, "B");
	break;
      case 1:
	strcpy(id, "kB");
	break;
      case 2:
	strcpy(id, "MB");
	break;
      case 3:
	strcpy(id, "GB");
	break;
      case 4:
	strcpy(id, "TB");
	break;
      default:
	fprintf(stderr, "connecttwo: a bug in number order!\n");
	strcpy(id, "Strange Unit");
	break;
      }
      sprintf(s, "bandwidth: %.2f %s/s", bw, id);
    }
    p->ot = nt;
    p->wbytes = 0;
    return 1;
  }
  return 0;
}

static int read_one(struct pipe *p)
{
  int ret;
  if (p->bytes == 0) {
    ret = read(p->src_fd, p->buf, p->bufsize);
    if (ret > 0) {
      p->bytes = ret;
    } else if (ret == 0) {
      p->eof_read = 1;
    } else {
      if (errno != EINTR) {
	perror("connecttwo: read error");
	p->eof_read = 1;
      }
    }
  } else {
    ret = write(p->dst_fd, &p->buf[p->offs], p->bytes - p->offs);
    if (ret > 0) {
      p->wbytes += ret;
      p->tbytes += ret;
      p->offs += ret;
    } else if (ret == 0) {
      fprintf(stderr, "%d: connecttwo: interesting: write returned 0\n", cpid);
      p->eof_read = 1;
      p->bytes = 0;
      p->offs = 0;
      return 0;
    } else if (ret < 0) {
      if (errno != EINTR) { 
	perror("connecttwo: write error");
	p->eof_read = 1;
	p->bytes = 0;
	p->offs = 0;
	return 0;
      }
    }
    if (p->offs == p->bytes) {
      p->bytes = p->offs = 0;
    }
  }

  return 1;
}

static void pipemeter_loop(struct pipe *p1, struct pipe *p2)
{
  const int aligned_size = 4096;
  char *aligned_buf;
  char *real_buf;
  struct sigaction act;
  struct timeval ot;
  int ret;
  char info[256];
  int valid_time = 1;
  int p1_active;
  int p2_active;

  struct pollfd pfds[4];
  int something_polled;
  int i;

  cpid = getpid();

#ifndef PAGE_SIZE
#define PAGE_SIZE 8192
#endif

  /* get page size aligned buffer of size 'aligned_size' */
  real_buf = malloc(aligned_size * 2 + PAGE_SIZE);
  if (!real_buf) {
    fprintf(stderr, "connecttwo %d: not enough memory\n", cpid);
    exit(-1);
  }
  aligned_buf = real_buf;
  aligned_buf += PAGE_SIZE - (((long) aligned_buf) & (PAGE_SIZE - 1));
  if (p1) {
    p1->buf = aligned_buf;
    p1->bufsize = aligned_size;
  }
  if (p2) {
    p2->buf = aligned_buf + aligned_size;
    p2->bufsize = aligned_size;
  }

  memset(&act, 0, sizeof(act));
  act.sa_handler = pipemeter_sigpipe;
  sigaction(SIGPIPE, &act, 0);

  if (gettimeofday(&ot, 0)) {
    memset(&ot, 0, sizeof(ot));
    perror ("connecttwo: gettimeofday() failed");
    valid_time = 0;
  }
  if (p1) {
    p1->ot = ot;
    p1->vot = ot;
  }
  if (p2) {
    p2->ot = ot;
    p2->vot = ot;
  }

  p1_active = (p1 != 0);
  p2_active = (p2 != 0);

  while (p1_active || p2_active) {

    if (p1->eof_read && p1->bytes == 0)
      p1_active = 0;

    if (p2->eof_read && p2->bytes == 0)
      p2_active = 0;

    if ((!p1_active || !p2_active) && p1->bytes == 0 && p2->bytes == 0) {
      break;
    }

    memset(pfds, 0, 4 * sizeof(struct pollfd));

    something_polled = 0;

    if (p1->eof_read == 0 && p1->bytes == 0) {
      pfds[0].fd = p1->src_fd;
      pfds[0].events = POLLIN;
      something_polled = 1;
    }
    if (p1->bytes) {
      pfds[1].fd = p1->dst_fd;
      pfds[1].events = POLLOUT;
      something_polled = 1;
    }

    if (p2->eof_read == 0 && p2->bytes == 0) {
      pfds[2].fd = p2->src_fd;
      pfds[2].events = POLLIN;
      something_polled = 1;
    }
    if (p2->bytes) {
      pfds[3].fd = p2->dst_fd;
      pfds[3].events = POLLOUT;
      something_polled = 1;
    }

    if (!something_polled)
      break;

    ret = poll(pfds, 4, -1);

    if (ret > 0) {
      int active;
      if (p1_active) {
	active = 0;
	for (i = 0; i < 2; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p1)) {
	    p1_active = 0;
	  }
	}
      }

      if (p2_active) {
	active = 0;
	for (i = 2; i < 4; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p2)) {
	    p2_active = 0;
	  }
	}
      }

    } else if (ret < 0) {
      if (errno != EINTR) {
	perror("poll error");
	goto out;
      }
    }

    if (valid_time) {
      if (p1_active && timetest(info, p1, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p1->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d ->: %s\n", cpid, info);
      }
      if (p2_active && timetest(info, p2, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p2->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d <-: %s\n", cpid, info);
      }
    }
  }

 out:

  if (p1) {
    p1->wbytes = p1->tbytes;
    p1->ot = p2->vot;
    timetest(info, p1, 1);
    fprintf(stderr, "%d ->: average %s\n", cpid, info);
    fprintf(stderr, "%d ->: total bytes: %lld\n", cpid, p1->tbytes);
  }
  if (p2) {
    p2->wbytes = p2->tbytes;
    p2->ot = p2->vot;
    timetest(info, p2, 1);
    fprintf(stderr, "%d <-: average %s\n", cpid, info);
    fprintf(stderr, "%d <-: total bytes: %lld\n", cpid, p2->tbytes);
  }

  free(real_buf);
}

static void pf_handle_client(int src_fd, int dst_fd)
{
  struct pipe p1;
  struct pipe p2;

  memset(&p1, 0, sizeof(struct pipe));
  memset(&p2, 0, sizeof(struct pipe));

  p1.src_fd = src_fd;
  p1.dst_fd = dst_fd;
  p2.src_fd = dst_fd;
  p2.dst_fd = src_fd;

  pipemeter_loop(&p1, &p2);

  close(p1.src_fd);
  close(p1.dst_fd);
  close(p2.src_fd);
  close(p2.dst_fd);
}

static int read_line(char *buf, int fd, int maxlen) {
  size_t bytes_read = 0;
  int ret;
  while (1) {
    if (bytes_read >= ((size_t) maxlen))
      return 0;
    ret = read(fd, &buf[bytes_read], 1);
    if (ret < 0) {
      if (errno == EAGAIN || errno == EINTR) {
        continue;
      }
      perror("connecttwo: could not read line");
      return 0;
    } else if (ret == 0) {
      fprintf(stderr, "connecttwo: unexpected eof when reading a line\n");
      buf[bytes_read] = 0;
      return 0;
    }
    if (buf[bytes_read] == '\n') {
      buf[bytes_read] = 0;
      break;
    }
    bytes_read += ret;
  }
  return 1;
}

static int connect_two(char *src_name, char *src_port, char *dst_name, char *dst_port)
{
  int src_fd;
  int dst_fd;
  pid_t child;
  int sleep_time;

  char line[256];

  sleep_time = 1;

  while (1) {

    if ((src_fd = net_open(src_name, src_port, "tcp", 0)) < 0) {
      fprintf(stderr, "connecttwo: couldn't connect to src: %s:%s. Sleeping %d seconds before re-try.\n", src_name, src_port, sleep_time);
      sleep(sleep_time);
      sleep_time *= 2;
      if (sleep_time > 64) {
	sleep_time = 64;
      }
      continue;
    }

    if (hello_from_src) {
      if (!read_line(line, src_fd, sizeof(line))) {
	fprintf(stderr, "connecttwo: didn't get hello from src. Sleeping %d seconds before re-try.\n", sleep_time);
	sleep(sleep_time);
	sleep_time *= 2;
	if (sleep_time > 64) {
	  sleep_time = 64;
	}
	continue;
      }
    }

    if ((dst_fd = net_open(dst_name, dst_port, "tcp", 0)) < 0) {
      fprintf(stderr, "connecttwo: couldn't connect to dst: %s:%s. Sleeping %d seconds before re-try.\n", dst_name, dst_port, sleep_time);
      sleep(sleep_time);
      sleep_time *= 2;
      if (sleep_time > 64) {
	sleep_time = 64;
      }
      continue;
    }
    sleep_time = 1;

    child = fork();
    if (!child) {
      child = fork();
      if (!child) {
	pf_handle_client(src_fd, dst_fd);
	fprintf(stderr, "connecttwo: client handled\n");
      } else if (child == -1) {
	perror("connecttwo: children can not reproduce");
      }
      abort();

    } else if (child == -1) {
      perror("connecttwo: can not have children. client will die.");
    } else {
      wait(0);
    }
    close(src_fd);
    close(dst_fd);
  }

  close(src_fd);
  close(dst_fd);
  fprintf(stderr, "connecttwo: this should not happen. main loop discontinued.\n");
  return 1;
}

static int get_host_port(char **address, char **port, char *in)
{
  char *sep;
  char *nstr;

  if (!(nstr = strdup(in)))
    return 0;

  if (!(sep = strrchr(nstr, (int) ':'))) {
    return 0;
  }

  *sep = 0;

  *address = strdup(nstr);
  *port = strdup(sep + 1);

  free (nstr);

  if (!(*address) || !(*port)) {
    free (*address);
    free (*port);
    return 0;
  }

  return 1;
}


int main(int argc, char **argv)
{
  int i;
  char *src_name = 0;
  char *src_port = 0;
  char *dst_name = 0;
  char *dst_port = 0;

  for (i = 1; i < argc;) {

    if (!strcmp("-s", argv[i])) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "not enough args. give source host:port\n");
        return -1;
      }
      if (!get_host_port(&src_name, &src_port, argv[i + 1])) {
        return -1;
      }
      i += 2;

    } else if (!strcmp("-d", argv[i])) {
      if ((i + 1) >= argc) {
        fprintf(stderr, "not enough args. give destination host:port\n");
        return -1;
      }
      if (!get_host_port(&dst_name, &dst_port, argv[i + 1])) {
        return -1;
      }
      i += 2;

    } else if (!strcmp("--hello-from-src", argv[i])) {
      hello_from_src = 1;
      i++;

    } else {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      return -1;
    }
  }

  if (!src_port) {
    fprintf(stderr, "you must provide source with -s host:port!\n");
    return -1;
  }
  if (!dst_port) {
    fprintf(stderr, "you must provide destination with -d host:port!\n");
    return -1;
  }

  return connect_two(src_name, src_port, dst_name, dst_port) ? -1 : 0;
}
