/* listen two port forward utility
Copyright (c) 2003 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"

extern int errno;

/* configuration variables */
static int default_interval = 2000;

static int hello_dst = 0;

/* per-client variables */
static pid_t cpid;

static void pipemeter_sigpipe(int sig)
{
  sig = sig;
  fprintf(stderr, "%d: listentwo: SIGPIPE caught. Killing client.\n", getpid());
}

struct pipe {
  int src_fd;
  int dst_fd;
  char *buf;
  int bufsize;
  int bytes;
  int offs;
  int eof_read;
  long long tbytes;
  long long wbytes;
  struct timeval ot;
  struct timeval vot;
};

static int timetest(char *s, struct pipe *p, int force)
{
  struct timeval nt;
  int t;
  strcpy(s, "bandwidth: NaN");
  if (gettimeofday(&nt, 0)) {
    /* time failed. bandwidth = NaN. return false. */
    return 0;
  }
  t = 1000 * (nt.tv_sec - p->ot.tv_sec) + ((int) nt.tv_usec)/1000 - ((int) p->ot.tv_usec)/1000;
  if (t < 0) {
    fprintf(stderr, "listentwo: chronoton particles detected. clock ran backwards. k3wl!\n");
    t = default_interval + 1;
  }
  if (t > default_interval || force) {
    int order;
    double bw;
    char id[256];
    if (t) {
      bw = ((double) p->wbytes) * 1000.0f / ((double) t);
      order = 0;
      while (bw >= 1024.0f) {
	bw /= 1024.0f;
	order++;
	if (order == 4)
	  break;
      }
      switch (order) {
      case 0:
	strcpy(id, "B");
	break;
      case 1:
	strcpy(id, "kB");
	break;
      case 2:
	strcpy(id, "MB");
	break;
      case 3:
	strcpy(id, "GB");
	break;
      case 4:
	strcpy(id, "TB");
	break;
      default:
	fprintf(stderr, "listentwo: a bug in number order!\n");
	strcpy(id, "Strange Unit");
	break;
      }
      sprintf(s, "bandwidth: %.2f %s/s", bw, id);
    }
    p->ot = nt;
    p->wbytes = 0;
    return 1;
  }
  return 0;
}

static int read_one(struct pipe *p)
{
  int ret;
  if (p->bytes == 0) {
    ret = read(p->src_fd, p->buf, p->bufsize);
    if (ret > 0) {
      p->bytes = ret;
    } else if (ret == 0) {
      p->eof_read = 1;
    } else {
      if (errno != EINTR) {
	perror("listentwo: read error");
	p->eof_read = 1;
      }
    }
  } else {
    ret = write(p->dst_fd, &p->buf[p->offs], p->bytes - p->offs);
    if (ret > 0) {
      p->wbytes += ret;
      p->tbytes += ret;
      p->offs += ret;
    } else if (ret == 0) {
      fprintf(stderr, "%d: listentwo: interesting: write returned 0\n", cpid);
      p->eof_read = 1;
      p->bytes = 0;
      p->offs = 0;
      return 0;
    } else if (ret < 0) {
      if (errno != EINTR) { 
	perror("listentwo: write error");
	p->eof_read = 1;
	p->bytes = 0;
	p->offs = 0;
	return 0;
      }
    }
    if (p->offs == p->bytes) {
      p->bytes = p->offs = 0;
    }
  }

  return 1;
}

static void pipemeter_loop(struct pipe *p1, struct pipe *p2)
{
  const int aligned_size = 4096;
  char *aligned_buf;
  char *real_buf;
  struct sigaction act;
  struct timeval ot;
  int ret;
  char info[256];
  int valid_time = 1;
  int p1_active;
  int p2_active;

  struct pollfd pfds[4];
  int something_polled;
  int i;

  cpid = getpid();

#ifndef PAGE_SIZE
#define PAGE_SIZE 8192
#endif

  /* get page size aligned buffer of size 'aligned_size' */
  real_buf = malloc(aligned_size * 2 + PAGE_SIZE);
  if (!real_buf) {
    fprintf(stderr, "listentwo %d: not enough memory\n", cpid);
    exit(-1);
  }
  aligned_buf = real_buf;
  aligned_buf += PAGE_SIZE - (((long) aligned_buf) & (PAGE_SIZE - 1));
  if (p1) {
    p1->buf = aligned_buf;
    p1->bufsize = aligned_size;
  }
  if (p2) {
    p2->buf = aligned_buf + aligned_size;
    p2->bufsize = aligned_size;
  }

  memset(&act, 0, sizeof(act));
  act.sa_handler = pipemeter_sigpipe;
  sigaction(SIGPIPE, &act, 0);

  if (gettimeofday(&ot, 0)) {
    memset(&ot, 0, sizeof(ot));
    perror ("listentwo: gettimeofday() failed");
    valid_time = 0;
  }
  if (p1) {
    p1->ot = ot;
    p1->vot = ot;
  }
  if (p2) {
    p2->ot = ot;
    p2->vot = ot;
  }

  p1_active = (p1 != 0);
  p2_active = (p2 != 0);

  while (p1_active || p2_active) {

    if (p1->eof_read && p1->bytes == 0)
      p1_active = 0;

    if (p2->eof_read && p2->bytes == 0)
      p2_active = 0;

    if ((!p1_active || !p2_active) && p1->bytes == 0 && p2->bytes == 0) {
      break;
    }

    memset(pfds, 0, 4 * sizeof(struct pollfd));

    something_polled = 0;

    if (p1->eof_read == 0 && p1->bytes == 0) {
      pfds[0].fd = p1->src_fd;
      pfds[0].events = POLLIN;
      something_polled = 1;
    }
    if (p1->bytes) {
      pfds[1].fd = p1->dst_fd;
      pfds[1].events = POLLOUT;
      something_polled = 1;
    }

    if (p2->eof_read == 0 && p2->bytes == 0) {
      pfds[2].fd = p2->src_fd;
      pfds[2].events = POLLIN;
      something_polled = 1;
    }
    if (p2->bytes) {
      pfds[3].fd = p2->dst_fd;
      pfds[3].events = POLLOUT;
      something_polled = 1;
    }

    if (!something_polled)
      break;

    ret = poll(pfds, 4, -1);

    if (ret > 0) {
      int active;
      if (p1_active) {
	active = 0;
	for (i = 0; i < 2; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p1)) {
	    p1_active = 0;
	  }
	}
      }

      if (p2_active) {
	active = 0;
	for (i = 2; i < 4; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p2)) {
	    p2_active = 0;
	  }
	}
      }

    } else if (ret < 0) {
      if (errno != EINTR) {
	perror("poll error");
	goto out;
      }
    }

    if (valid_time) {
      if (p1_active && timetest(info, p1, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p1->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d ->: %s\n", cpid, info);
      }
      if (p2_active && timetest(info, p2, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p2->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d <-: %s\n", cpid, info);
      }
    }
  }

 out:

  if (p1) {
    p1->wbytes = p1->tbytes;
    p1->ot = p2->vot;
    timetest(info, p1, 1);
    fprintf(stderr, "%d ->: average %s\n", cpid, info);
    fprintf(stderr, "%d ->: total bytes: %lld\n", cpid, p1->tbytes);
  }
  if (p2) {
    p2->wbytes = p2->tbytes;
    p2->ot = p2->vot;
    timetest(info, p2, 1);
    fprintf(stderr, "%d <-: average %s\n", cpid, info);
    fprintf(stderr, "%d <-: total bytes: %lld\n", cpid, p2->tbytes);
  }

  free(real_buf);
}

static void pf_handle_client(int src_fd, int dst_fd)
{
  struct pipe p1;
  struct pipe p2;

  memset(&p1, 0, sizeof(struct pipe));
  memset(&p2, 0, sizeof(struct pipe));

  p1.src_fd = src_fd;
  p1.dst_fd = dst_fd;
  p2.src_fd = dst_fd;
  p2.dst_fd = src_fd;

  pipemeter_loop(&p1, &p2);

  close(p1.src_fd);
  close(p1.dst_fd);
  close(p2.src_fd);
  close(p2.dst_fd);
}

static ssize_t atomic_write(int fd, const void *buf, size_t count)
{
  char *b = (char *) buf;
  size_t bytes_written = 0;
  int ret;
  while (bytes_written < count) {
    ret = write(fd, &b[bytes_written], count - bytes_written);
    if (ret < 0) {
      if (errno == EINTR) {
        continue;
      }
      if (bytes_written == 0)
        bytes_written = -1;
      break;
    }
    bytes_written += ret;
  }
  return bytes_written;
}

static int listen_two(char *src_port, char *dst_port)
{
  int server_fd1;
  int server_fd2;
  int src_fd;
  int dst_fd;
  pid_t child;

  int i;
  int state;
  int ret;
  int len;

  struct pollfd pfds[2];
  int npfds;

  char sockbuf[256];
  struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
  int socklen;
  struct net_host_id id;
  char printed_fd1_id[256];
  char printed_fd2_id[256];

  if ((server_fd1 = net_bind(0, 0, src_port, "tcp", 1 /* REUSE */)) < 0) {
    fprintf(stderr, "listentwo: can't bind server port\n");
    exit(-1);
  }
  if (listen(server_fd1, 5)) {
    perror("listentwo: can't listen on src port");
    exit(-1);
  }

  if ((server_fd2 = net_bind(0, 0, dst_port, "tcp", 1 /* REUSE */)) < 0) {
    fprintf(stderr, "listentwo: can't bind dst port\n");
    exit(-1);
  }
  if (listen(server_fd2, 5)) {
    perror("listentwo: can't listen on src port");
    exit(-1);
  }

  state = 0;

  while (1) {

    if (state != 3) {

      npfds = 0;
      if ((state & 1) == 0) {
	pfds[npfds].fd = server_fd1;
	pfds[npfds].events = POLLIN;
	npfds++;
      }
      if ((state & 2) == 0) {
	pfds[npfds].fd = server_fd2;
	pfds[npfds].events = POLLIN;
	npfds++;
      }

      ret = poll(pfds, npfds, -1);

      if (ret == 0) {
	continue;
      } else if (ret < 0) {
	if (errno == EINTR) {
	  continue;
	}
	goto err_out;
      }

      for (i = 0; i < npfds; i++) {
	if (pfds[i].revents & POLLIN) {
	  if (pfds[i].fd == server_fd1) {
	    socklen = sizeof(sockbuf);
	    src_fd = accept(server_fd1, (struct sockaddr *) sockname, &socklen);
	    if (src_fd < 0) {
	      fprintf(stderr, "server_fd1 connection failed\n");
	      continue;
	    }
	    if (!net_get_id(&id, (struct sockaddr *) sockname)) {
	      fprintf(stderr, "client connected to server_fd1. can't get info on socket. rejecting.\n");
	      close(src_fd);
	    }
	    /* append string with connecting ip to printed_id */
	    printed_fd1_id[0] = 0;
	    net_print_id(printed_fd1_id, &id, sizeof(printed_fd1_id));
	    fprintf(stderr, "server_fd1: %s connected\n", printed_fd1_id);
	    state |= 1;

	  } else {
	    socklen = sizeof(sockbuf);
	    dst_fd = accept(server_fd2, (struct sockaddr *) sockname, &socklen);
	    if (dst_fd < 0) {
	      fprintf(stderr, "server_fd2 connection failed\n");
	      continue;
	    }
	    if (!net_get_id(&id, (struct sockaddr *) sockname)) {
	      fprintf(stderr, "client connected to server_fd1. can't get info on socket. rejecting.\n");
	      close(dst_fd);
	    }
	    /* append string with connecting ip to printed_id */
	    printed_fd2_id[0] = 0;
	    net_print_id(printed_fd2_id, &id, sizeof(printed_fd2_id));
	    fprintf(stderr, "server_fd2: %s connected\n", printed_fd2_id);
	    state |= 2;
	  }
	}
      }
      continue;

    } else {

      child = fork();
      if (!child) {
	child = fork();
	if (!child) {

	  if (hello_dst) {
	    len = strlen(printed_fd1_id);
	    if (atomic_write(dst_fd, printed_fd1_id, len) < len) {
	      fprintf(stderr, "listentwo: couldn't say hello: %s -> %s\n", printed_fd1_id, printed_fd2_id);
	      abort();
	    }
	    len = 1;
	    if (atomic_write(dst_fd, "\n", len) < len) {
	      fprintf(stderr, "listentwo: couldn't say hello: %s -> %s\n", printed_fd1_id, printed_fd2_id);
	      abort();
	    }
	  }

	  pf_handle_client(src_fd, dst_fd);
	  fprintf(stderr, "client handled: %s -> %s\n", printed_fd1_id, printed_fd2_id);
	} else if (child == -1) {
	  perror("children can not reproduce");
	}
	abort();

      } else if (child == -1) {
	perror("can not have children. client will die.");
      } else {
	wait(0);
      }
      close(src_fd);
      close(dst_fd);
      state = 0;
    }
  }

  fprintf(stderr, "this should not happen. main loop discontinued.\n");
  return 1;

 err_out:
  close(server_fd1);
  close(server_fd2);
  return 0;
}


int main(int argc, char **argv)
{
  int i;
  char *src_port = 0;
  char *dst_port = 0;

  for (i = 1; i < argc;) {

    if (!strcmp("-s", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give source port.\n");
	return -1;
      }
      if (!(src_port = strdup(argv[i + 1]))) {
	fprintf(stderr, "not enough memory for src_port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-d", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give destination port.\n");
	return -1;
      }
      if (!(dst_port = strdup(argv[i + 1]))) {
	fprintf(stderr, "not enough memory for dst_port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("--hello-dst", argv[i])) {
      hello_dst = 1;
      i++;

    } else {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      return -1;
    }
  }

  if (!src_port) {
    fprintf(stderr, "you must provide source port with -s port!\n");
    return -1;
  }
  if (!dst_port) {
    fprintf(stderr, "you must provide destination port with -d port!\n");
    return -1;
  }

  return listen_two(src_port, dst_port) ? -1 : 0;
}
