/* tcp port forward utility
Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#define _XOPEN_SOURCE 600
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"
#include "accounting.h"

struct net_host_id *allowed_hosts = NULL;
int n_allowed_hosts = 0;

/* configuration variables */
static int default_interval = 2000;
static int kill_on_eof = 1;

/* per-client variables */
static pid_t cpid;

static char *accounting_file = NULL;

static void pipemeter_sigpipe(int sig)
{
  sig = sig;
  fprintf(stderr, "%d: tcp-pf: SIGPIPE caught. Killing client.\n", getpid());
}

struct pipe {
  int src_fd;
  int dst_fd;
  char *buf;
  int bufsize;
  int bytes;
  int offs;
  int eof_read;
  long long tbytes;
  long long wbytes;
  struct timeval ot;
  struct timeval vot;
};

static int timetest(char *s, struct pipe *p, int force)
{
  struct timeval nt;
  int t;
  strcpy(s, "bandwidth: NaN");
  if (gettimeofday(&nt, 0)) {
    /* time failed. bandwidth = NaN. return false. */
    return 0;
  }
  t = 1000 * (nt.tv_sec - p->ot.tv_sec) + ((int) nt.tv_usec)/1000 - ((int) p->ot.tv_usec)/1000;
  if (t < 0) {
    fprintf(stderr, "tcp-pf: chronoton particles detected. clock ran backwards. k3wl!\n");
    t = default_interval + 1;
  }
  if (t > default_interval || force) {
    int order;
    double bw;
    char id[256];
    if (t) {
      bw = ((double) p->wbytes) * 1000.0f / ((double) t);
      order = 0;
      while (bw >= 1024.0f) {
	bw /= 1024.0f;
	order++;
	if (order == 4)
	  break;
      }
      switch (order) {
      case 0:
	strcpy(id, "B");
	break;
      case 1:
	strcpy(id, "KiB");
	break;
      case 2:
	strcpy(id, "MiB");
	break;
      case 3:
	strcpy(id, "GiB");
	break;
      case 4:
	strcpy(id, "TiB");
	break;
      default:
	fprintf(stderr, "tcp-pf: a bug in number order!\n");
	strcpy(id, "Strange Unit");
	break;
      }
      sprintf(s, "bandwidth: %.2f %s/s", bw, id);
    }
    p->ot = nt;
    p->wbytes = 0;
    return 1;
  }
  return 0;
}

static int read_one(struct pipe *p)
{
  int ret;
  if (p->bytes == 0) {
    ret = read(p->src_fd, p->buf, p->bufsize);
    if (ret > 0) {
      p->bytes = ret;
    } else if (ret == 0) {
      p->eof_read = 1;
    } else {
      if (errno != EINTR) {
	perror("tcp-pf: read error");
	p->eof_read = 1;
      }
    }
  } else {
    ret = write(p->dst_fd, &p->buf[p->offs], p->bytes - p->offs);
    if (ret > 0) {
      p->wbytes += ret;
      p->tbytes += ret;
      p->offs += ret;
    } else if (ret == 0) {
      fprintf(stderr, "%d: tcp-pf: interesting: write returned 0\n", cpid);
      p->eof_read = 1;
      p->bytes = 0;
      p->offs = 0;
      return 0;
    } else if (ret < 0) {
      if (errno != EINTR) { 
	perror("tcp-pf: write error");
	p->eof_read = 1;
	p->bytes = 0;
	p->offs = 0;
	return 0;
      }
    }
    if (p->offs == p->bytes) {
      p->bytes = p->offs = 0;
    }
  }

  return 1;
}

static void pipemeter_loop(struct pipe *p1, struct pipe *p2)
{
  const int aligned_size = 8192;
  char *aligned_buf;
  void *real_buf;
  struct sigaction act;
  struct timeval ot;
  int ret;
  char info[256];
  int valid_time = 1;
  int p1_active;
  int p2_active;

  struct pollfd pfds[4];
  int something_polled;
  int i;

  cpid = getpid();

  if (posix_memalign(&real_buf, 8192, aligned_size)) {
    fprintf(stderr, "tcp-pf %d: not enough memory: %s\n", cpid, strerror(errno));
    exit(-1);
  }
  aligned_buf = real_buf;

  if (p1) {
    p1->buf = aligned_buf;
    p1->bufsize = aligned_size;
  }
  if (p2) {
    p2->buf = aligned_buf + aligned_size;
    p2->bufsize = aligned_size;
  }

  memset(&act, 0, sizeof(act));
  act.sa_handler = pipemeter_sigpipe;
  sigaction(SIGPIPE, &act, 0);

  if (gettimeofday(&ot, 0)) {
    memset(&ot, 0, sizeof(ot));
    perror ("tcp-pf: gettimeofday() failed");
    valid_time = 0;
  }
  if (p1) {
    p1->ot = ot;
    p1->vot = ot;
  }
  if (p2) {
    p2->ot = ot;
    p2->vot = ot;
  }

  p1_active = (p1 != 0);
  p2_active = (p2 != 0);

  while (p1_active || p2_active) {

    if (p1->eof_read && p1->bytes == 0)
      p1_active = 0;

    if (p2->eof_read && p2->bytes == 0)
      p2_active = 0;

    if (kill_on_eof) {
      if ((!p1_active || !p2_active) && p1->bytes == 0 && p2->bytes == 0) {
	break;
      }
    }


    memset(pfds, 0, 4 * sizeof(struct pollfd));

    something_polled = 0;

    if (p1->eof_read == 0 && p1->bytes == 0) {
      pfds[0].fd = p1->src_fd;
      pfds[0].events = POLLIN;
      something_polled = 1;
    }
    if (p1->bytes) {
      pfds[1].fd = p1->dst_fd;
      pfds[1].events = POLLOUT;
      something_polled = 1;
    }

    if (p2->eof_read == 0 && p2->bytes == 0) {
      pfds[2].fd = p2->src_fd;
      pfds[2].events = POLLIN;
      something_polled = 1;
    }
    if (p2->bytes) {
      pfds[3].fd = p2->dst_fd;
      pfds[3].events = POLLOUT;
      something_polled = 1;
    }

    if (!something_polled)
      break;

    ret = poll(pfds, 4, -1);

    if (ret > 0) {
      int active;
      if (p1_active) {
	active = 0;
	for (i = 0; i < 2; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p1)) {
	    p1_active = 0;
	  }
	}
      }

      if (p2_active) {
	active = 0;
	for (i = 2; i < 4; i++) {
	  if (pfds[i].revents & (POLLERR | POLLHUP)) {
	    active = 1;
	  }
	  if (pfds[i].revents & pfds[i].events)
	    active = 1;
	}
	if (active) {
	  if (!read_one(p2)) {
	    p2_active = 0;
	  }
	}
      }

    } else if (ret < 0) {
      if (errno != EINTR) {
	perror("poll error");
	goto out;
      }
    }

    if (valid_time) {
      if (p1_active && timetest(info, p1, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p1->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d ->: %s\n", cpid, info);
      }
      if (p2_active && timetest(info, p2, 0)) {
	char byte_info[256];
	sprintf(byte_info, "\tbytes: %lld", p2->tbytes);
	strcat(info, byte_info);
	fprintf(stderr, "%d <-: %s\n", cpid, info);
      }
    }
  }

 out:

  if (p1) {
    p1->wbytes = p1->tbytes;
    p1->ot = p1->vot;
    timetest(info, p1, 1);
    fprintf(stderr, "%d ->: average %s\n", cpid, info);
    fprintf(stderr, "%d ->: total bytes: %lld\n", cpid, p1->tbytes);
  }
  if (p2) {
    p2->wbytes = p2->tbytes;
    p2->ot = p2->vot;
    timetest(info, p2, 1);
    fprintf(stderr, "%d <-: average %s\n", cpid, info);
    fprintf(stderr, "%d <-: total bytes: %lld\n", cpid, p2->tbytes);
  }

  free(real_buf);
}

static int get_host_port(char **address, char **port, char *in)
{
  char *sep;
  char *nstr;

  if (!(nstr = strdup(in)))
    return 0;

  if (!(sep = strrchr(nstr, (int) ':'))) {
    return 0;
  }

  *sep = 0;

  *address = strdup(nstr);
  *port = strdup(sep + 1);

  free (nstr);

  if (!(*address) || !(*port)) {
    free (*address);
    free (*port);
    return 0;
  }

  return 1;
}

static void pf_handle_client(int src_fd, char *dst_host, char *dst_port)
{
  int dst_fd;
  struct pipe p1;
  struct pipe p2;
  char buf[1024];

  if (accounting_file) {
    get_account(buf, sizeof(buf), accounting_file);
    dst_host = buf;
  }

  if ((dst_fd = net_open(dst_host, dst_port, "tcp", 0)) < 0) {
    fprintf(stderr, "tcp-pf: couldn't connect to %s:%s\n", dst_host, dst_port);
    return;
  }

  memset(&p1, 0, sizeof(struct pipe));
  memset(&p2, 0, sizeof(struct pipe));

  p1.src_fd = src_fd;
  p1.dst_fd = dst_fd;
  p2.src_fd = dst_fd;
  p2.dst_fd = src_fd;

  pipemeter_loop(&p1, &p2);

  if (accounting_file != NULL)
    account_increment(dst_host, accounting_file, p1.tbytes + p2.tbytes);

  close(p1.src_fd);
  close(p1.dst_fd);
  close(p2.src_fd);
  close(p2.dst_fd);
}

static int pf_loop(char *src_port, char *dst_host, char *dst_port)
{
  int server_fd;
  int src_fd;
  pid_t child;

  char sockbuf[256];
  struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
  int socklen;
  int i;
  struct net_host_id id;
  char printed_id[256];
  int got_info;

  if ((server_fd = net_bind(0, 0, src_port, "tcp", 1 /* REUSE */)) < 0) {
    fprintf(stderr, "tcp-pf: can't bind server port\n");
    exit(-1);
  }

  if (listen(server_fd, 5)) {
    perror("can't listen on server port");
    exit(-1);
  }

  while (1) {
    socklen = sizeof(sockbuf);
    src_fd = accept(server_fd, (struct sockaddr *) sockname, &socklen);
    if (src_fd < 0) {
      if (errno == EAGAIN || errno == EINTR) {
	continue;
      }
      perror("client accept failed");
      goto err_out;
    }

    if (!(got_info = net_get_id(&id, (struct sockaddr *) sockname))) {
      fprintf(stderr, "client connected. can't get info on socket. rejecting.\n");
      if (n_allowed_hosts > 0) {
	goto reject;
      }
    }

    if (got_info) {
      printed_id[0] = 0;
      net_print_id(printed_id, &id, sizeof(printed_id));

      if (n_allowed_hosts > 0) {
	for (i = 0; i < n_allowed_hosts; i++) {
	  if (net_are_same_ips(&id, allowed_hosts + i))
	    goto accept;
	}
	fprintf(stderr, "client connected from %s => reject\n", printed_id);
	goto reject;
      }
    accept:
      fprintf(stderr, "client connected from %s => accept\n", printed_id);
    }

    child = fork();
    if (!child) {
      close(server_fd);
      child = fork();
      if (!child) {
	pf_handle_client(src_fd, dst_host, dst_port);
	printf("client handled\n");
      } else if (child == -1) {
	perror("children can not reproduce");
      }
      abort();

    } else if (child == -1) {
      perror("can not have children. client will die.");

    } else {
      wait(0);
    }

  reject:

    close(src_fd);
  }

  return 1;

 err_out:
  close(server_fd);
  return 0;
}

static void print_help(void)
{
  printf("\n");
  printf("tcp-pf %s by Heikki Orsila <heikki.orsila@iki.fi>\n", VERSION);
  printf("\n");
  printf("OPTIONS\n");
  printf("\n");
  printf(" -a/--allow host   allow incoming connections from host. this option can be\n");
  printf("                   used many times to select many hosts\n");
  printf(" -d host:port      forward incoming connections to host:port\n");
  printf(" -h/--help         this crap\n");
  printf(" --lazy            let tcp connection hang until both ends have shutdown\n");
  printf(" -l/--list file    multiplex target connections to hosts listed in file\n");
  printf("                   according to transfer history. the host that has least\n");
  printf("                   transfered bytes is selected. (network load balancing)\n");
  printf(" -p/--dport port   set destination port (use with -l)\n");
  printf(" -s port           set source port to accept connections\n");
}


int main(int argc, char **argv)
{
  int i;
  char *src_port = NULL;
  char *dst_host = NULL;
  char *dst_port = NULL;
  struct net_host_id id;
  struct net_host_id *new_allowed;

  for (i = 1; i < argc;) {

    if (!strcmp("-s", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give source port.\n");
	print_help();
	return -1;
      }
      if (!(src_port = strdup(argv[i + 1]))) {
	fprintf(stderr, "not enough memory for src_port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-d", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give destination host:port\n");
	print_help();
	return -1;
      }
      if (!get_host_port(&dst_host, &dst_port, argv[i + 1])) {
	return -1;
      }
      i += 2;

    } else if (!strcmp("--list", argv[i]) || !strcmp("-l", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give accounting file\n");
	print_help();
	return -1;
      }
      accounting_file = strdup(argv[i + 1]);
      if (accounting_file == NULL) {
	fprintf(stderr, "not enough memory for accounting filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("--dport", argv[i]) || !strcmp("-p", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give dst port.\n");
	print_help();
	return -1;
      }
      if ((dst_port = strdup(argv[i + 1])) == NULL) {
	fprintf(stderr, "not enough memory for dst port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("--lazy", argv[i])) {
      kill_on_eof = 0;
      i++;

    } else if (!strcmp("--allow", argv[i]) || !strcmp("-a", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give allowed hostname.\n");
	print_help();
	return -1;
      }
      if (!net_info(&id, argv[i + 1])) {
	fprintf(stderr, "nothing known about %s\n", argv[i + 1]);
	return -1;
      }
      new_allowed = realloc(allowed_hosts, (n_allowed_hosts + 1) * sizeof(struct net_host_id));
      if (!new_allowed) {
	fprintf(stderr, "no memory for allowed hosts\n");
	return -1;
      }
      allowed_hosts = new_allowed;
      allowed_hosts[n_allowed_hosts] = id;
      n_allowed_hosts++;
      i += 2;

    } else if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      print_help();
      return 0;

    } else {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      print_help();
      return -1;
    }
  }

  if (src_port == NULL) {
    fprintf(stderr, "you must provide source port with -s port!\n");
    print_help();
    return -1;
  }

  if (accounting_file == NULL) {
    if (dst_host == NULL || dst_port == NULL) {
      fprintf(stderr, "you must provide destination host:port with -d host:port, or alternatively\n");
      fprintf(stderr, "you can provide destination list and destination port with -l listfile -p port\n");
      print_help();
      return -1;
    }
  } else if (dst_port == NULL) {
    fprintf(stderr, "you must provide destinatio port with -p port\n");
    print_help();
    return -1;
  }

  return pf_loop(src_port, dst_host, dst_port) ? -1 : 0;
}
